from collections.abc import Iterable

from pgai.semantic_catalog.models import (
    Fact,
    Procedure,
    SQLExample,
    Table,
    View,
)
from pgai.semantic_catalog.templates import env


def render_table(table: Table) -> str:
    """Render a table object using the table template.

    Args:
        table: The Table object to render.

    Returns:
        The rendered table as a string.
    """
    template = env.get_template("table.j2")
    return template.render(table=table)


def render_tables(tables: Iterable[Table]) -> str:
    """Render multiple table objects.

    Args:
        tables: An iterable of Table objects to render.

    Returns:
        A string containing all rendered tables separated by newlines.
    """
    return "\n".join(map(render_table, tables)).strip()


def render_view(view: View) -> str:
    """Render a view object using the view template.

    Args:
        view: The View object to render.

    Returns:
        The rendered view as a string.
    """
    template = env.get_template("view.j2")
    return template.render(view=view)


def render_views(views: Iterable[View]) -> str:
    """Render multiple view objects.

    Args:
        views: An iterable of View objects to render.

    Returns:
        A string containing all rendered views separated by newlines.
    """
    return "\n".join(map(render_view, views)).strip()


def render_procedure(proc: Procedure) -> str:
    """Render a procedure object using the procedure template.

    Args:
        proc: The Procedure object to render.

    Returns:
        The rendered procedure as a string.
    """
    template = env.get_template("procedure.j2")
    return template.render(proc=proc)


def render_procedures(procedures: Iterable[Procedure]) -> str:
    """Render multiple procedure objects.

    Args:
        procedures: An iterable of Procedure objects to render.

    Returns:
        A string containing all rendered procedures separated by newlines.
    """
    return "\n".join(map(render_procedure, procedures)).strip()


def render_object(object: Table | View | Procedure) -> str:
    """Render a database object based on its type.

    Args:
        object: A database object (Table, View, or Procedure) to render.

    Returns:
        The rendered object as a string.
    """
    match object:
        case Table():
            return render_table(object)
        case View():
            return render_view(object)
        case Procedure():
            return render_procedure(object)


def render_objects(objects: Iterable[Table | View | Procedure]) -> str:
    """Render multiple database objects of various types.

    Args:
        objects: An iterable of database objects (Tables, Views, or Procedures) to render.

    Returns:
        A string containing all rendered objects separated by newlines.
    """
    return "\n".join(map(render_object, objects)).strip()


def render_fact(fact: Fact) -> str:
    """Render a fact object using the fact template.

    Args:
        fact: The Fact object to render.

    Returns:
        The rendered fact as a string.
    """
    template = env.get_template("fact.j2")
    return template.render(fact=fact)


def render_facts(facts: Iterable[Fact]) -> str:
    """Render multiple fact objects.

    Args:
        facts: An iterable of Fact objects to render.

    Returns:
        A string containing all rendered facts separated by newlines.
    """
    return "\n".join(map(render_fact, facts)).strip()


def render_sql_example(example: SQLExample) -> str:
    """Render a SQL example object using the sql_example template.

    Args:
        example: The SQLExample object to render.

    Returns:
        The rendered SQL example as a string.
    """
    template = env.get_template("sql_example.j2")
    return template.render(example=example)


def render_sql_examples(examples: Iterable[SQLExample]) -> str:
    """Render multiple SQL example objects.

    Args:
        examples: An iterable of SQLExample objects to render.

    Returns:
        A string containing all rendered SQL examples separated by newlines.
    """
    return "\n".join(map(render_sql_example, examples)).strip()



import asyncio

import psycopg
import pytest

import pgai
from pgai.semantic_catalog import semantic_catalog
from tests.semantic_catalog.utils import PostgresContainer

DATABASE = "sc_" + __name__.split(".")[-1]


@pytest.fixture(scope="module")
def event_loop():
    loop = asyncio.get_event_loop_policy().new_event_loop()
    yield loop
    loop.close()


@pytest.fixture(scope="module", autouse=True)
async def setup_database(container: PostgresContainer) -> None:
    container.drop_database(DATABASE, force=True)
    container.create_database(DATABASE)
    pgai.install(container.connection_string(database=DATABASE))
    async with await psycopg.AsyncConnection.connect(
        container.connection_string(database=DATABASE)
    ) as con:
        await semantic_catalog.create(con)


async def test_object(container: PostgresContainer) -> None:
    async with await psycopg.AsyncConnection.connect(
        container.connection_string(database=DATABASE)
    ) as con:
        sc = await semantic_catalog.from_name(con, "default")
        obj = await sc.add_object(
            con,
            1234,
            5678,
            1256,
            "table",
            ["public", "test_object"],
            [],
            "This is a test object",
        )
        assert obj is not None
        assert obj.id is not None
        assert obj.classid == 1234
        assert obj.objid == 5678
        assert obj.objsubid == 1256
        assert obj.objtype == "table"
        assert obj.objnames == ["public", "test_object"]
        assert obj.objargs == []
        assert obj.description == "This is a test object"

        objects = await sc.list_objects(con)
        assert len(objects) == 1

        edit_obj = await sc.update_object(con, obj.id, "This is an updated test object")
        assert edit_obj is not None
        assert edit_obj.id == obj.id
        assert edit_obj.description == "This is an updated test object"

        await sc.drop_object(con, obj.id)

        objects = await sc.list_objects(con)
        assert len(objects) == 0


async def test_object_not_found_errors(container: PostgresContainer) -> None:
    async with await psycopg.AsyncConnection.connect(
        container.connection_string(database=DATABASE)
    ) as con:
        sc = await semantic_catalog.from_name(con, "default")
        with pytest.raises(RuntimeError):
            await sc.update_object(
                con, 9999, "Object with id 9999 not found in catalog default"
            )


async def test_sql_example(container: PostgresContainer) -> None:
    async with await psycopg.AsyncConnection.connect(
        container.connection_string(database=DATABASE)
    ) as con:
        sc = await semantic_catalog.from_name(con, "default")
        sql_example = await sc.add_sql_example(
            con,
            "SELECT 1",
            "This is a test SQL example",
        )
        assert sql_example is not None
        assert sql_example.id is not None
        assert sql_example.description == "This is a test SQL example"
        assert sql_example.sql == "SELECT 1"

        sql_examples = await sc.list_sql_examples(con)
        assert len(sql_examples) == 1

        edit_sql_example = await sc.update_sql_example(
            con,
            sql_example.id,
            "SELECT 2",
        )
        assert edit_sql_example is not None
        assert edit_sql_example.id == sql_example.id
        assert edit_sql_example.description == "This is a test SQL example"
        assert edit_sql_example.sql == "SELECT 2"

        edit_sql_example = await sc.update_sql_example(
            con, sql_example.id, description="This is an updated test SQL example"
        )
        assert edit_sql_example is not None
        assert edit_sql_example.id == sql_example.id
        assert edit_sql_example.description == "This is an updated test SQL example"
        assert edit_sql_example.sql == "SELECT 2"

        await sc.drop_sql_example(con, sql_example.id)
        sql_examples = await sc.list_sql_examples(con)
        assert len(sql_examples) == 0


async def test_sql_example_not_found_errors(container: PostgresContainer) -> None:
    async with await psycopg.AsyncConnection.connect(
        container.connection_string(database=DATABASE)
    ) as con:
        sc = await semantic_catalog.from_name(con, "default")
        with pytest.raises(RuntimeError):
            await sc.update_sql_example(con, 9999, "SELECT 1")


async def test_fact(container: PostgresContainer) -> None:
    async with await psycopg.AsyncConnection.connect(
        container.connection_string(database=DATABASE)
    ) as con:
        sc = await semantic_catalog.from_name(con, "default")
        fact = await sc.add_fact(
            con,
            "This is a test fact",
        )
        assert fact is not None
        assert fact.id is not None
        assert fact.description == "This is a test fact"

        facts = await sc.list_facts(con)
        assert len(facts) == 1

        edit_fact = await sc.update_fact(con, fact.id, "This is an updated test fact")
        assert edit_fact is not None
        assert edit_fact.id == fact.id
        assert edit_fact.description == "This is an updated test fact"

        await sc.drop_fact(con, fact.id)

        facts = await sc.list_facts(con)
        assert len(facts) == 0


async def test_update_nonexistant_fact_errors(container: PostgresContainer) -> None:
    async with await psycopg.AsyncConnection.connect(
        container.connection_string(database=DATABASE)
    ) as con:
        sc = await semantic_catalog.from_name(con, "default")
        with pytest.raises(RuntimeError):
            await sc.update_fact(
                con, 9999, "Fact with id 9999 not found in catalog default"
            )



import os
from dataclasses import dataclass
from io import BytesIO
from typing import Any, Literal

from filetype import filetype  # type: ignore
from pydantic import BaseModel


@dataclass
class LoadedDocument:
    content: BytesIO
    file_path: str | None = None
    file_type: str | None = None


def guess_filetype(file_like: BytesIO, file_path: str | None = None) -> str | None:
    guess = filetype.guess(file_like)  # type: ignore
    file_like.seek(0)
    if guess is None:
        if file_path is None:
            return None
        return file_path.split(".")[-1]
    return guess.extension


class ColumnLoading(BaseModel):
    implementation: Literal["column"]
    column_name: str
    retries: int = 6

    def load(self, row: dict[str, str]) -> str | LoadedDocument:
        content = row[self.column_name] or ""
        if isinstance(content, bytes):
            content = BytesIO(content)
            return LoadedDocument(
                content=content,
                file_type=guess_filetype(content),
            )
        return content


class UriLoading(BaseModel):
    implementation: Literal["uri"]
    column_name: str
    retries: int = 6
    aws_role_arn: str | None = None

    def load(self, row: dict[str, str]) -> LoadedDocument:
        # Note: deferred import to avoid import overhead
        import boto3
        import smart_open  # type: ignore
        from mypy_boto3_s3.client import S3Client
        from mypy_boto3_sts.client import STSClient

        file_path = row[self.column_name]

        transport_params = None
        if file_path.startswith("s3://") and self.aws_role_arn is not None:
            external_id = os.getenv("AWS_ASSUME_ROLE_EXTERNAL_ID")
            sts_client: STSClient = boto3.client("sts")  # type: ignore
            kwargs: dict[str, Any] = {}
            if external_id is not None:
                kwargs["ExternalId"] = external_id
            assumed_role = sts_client.assume_role(
                RoleArn=self.aws_role_arn,
                RoleSessionName="timescale-vectorizer",
                **kwargs,
            )
            # Extract the temporary credentials
            credentials = assumed_role["Credentials"]

            # Create a boto3 session with the assumed role credentials
            session = boto3.Session(
                aws_access_key_id=credentials["AccessKeyId"],
                aws_secret_access_key=credentials["SecretAccessKey"],
                aws_session_token=credentials["SessionToken"],
            )

            # Create an S3 client using the session with assumed role
            s3_client: S3Client = session.client("s3")  # type: ignore
            transport_params = {"client": s3_client}
        content = BytesIO(
            smart_open.open(  # type: ignore
                file_path, "rb", transport_params=transport_params
            ).read()
        )
        return LoadedDocument(
            content=content,
            file_path=file_path,
            file_type=guess_filetype(content, file_path),
        )


class LoadingError(Exception):
    """
    Raised when the loader fails.
    """

    def __init__(self, *args: str, e: Exception):
        super().__init__(*args)
        self.__cause__ = e

    msg = "loading failed"



import psycopg

from pgai.semantic_catalog import describe, loader

from .utils import PostgresContainer, get_procedure_dict, get_table_dict, get_view_dict


async def test_load_tables(container: PostgresContainer):
    async with await psycopg.AsyncConnection.connect(
        container.connection_string(database="postgres_air")
    ) as con:
        for table_name, expected in get_table_dict().items():
            oids = await describe.find_tables(con, include_table=f"^{table_name}$")
            actual = await loader.load_tables(con, oids)
            assert len(actual) == 1
            actual = actual[0]
            # classid and objid will change. don't compare
            actual.classid = expected.classid
            actual.objid = expected.objid
            if actual.columns:
                for column in actual.columns:
                    column.classid = expected.classid
                    column.objid = expected.objid
            assert actual == expected, f"load_tables failed for {table_name}"


async def test_load_views(container: PostgresContainer):
    async with await psycopg.AsyncConnection.connect(
        container.connection_string(database="postgres_air")
    ) as con:
        for view_name, expected in get_view_dict().items():
            oids = await describe.find_views(con, include_view=f"^{view_name}$")
            assert len(oids) == 1
            actual = await loader.load_views(con, oids)
            assert len(actual) == 1
            actual = actual[0]
            # classid and objid will change. don't compare
            actual.classid = expected.classid
            actual.objid = expected.objid
            if actual.columns:
                for column in actual.columns:
                    column.classid = expected.classid
                    column.objid = expected.objid
            assert actual == expected, f"load_views failed for {view_name}"


async def test_load_procedures(container: PostgresContainer):
    async with await psycopg.AsyncConnection.connect(
        container.connection_string(database="postgres_air")
    ) as con:
        for proc_name, expected in get_procedure_dict().items():
            oids = await describe.find_procedures(con, include_proc=f"^{proc_name}$")
            assert len(oids) == 1, f"no oids found for {proc_name}"
            actual = await loader.load_procedures(con, oids)
            assert len(actual) == 1, f"no procedure found for {proc_name}"
            actual = actual[0]
            # classid and objid will change. don't compare
            actual.classid = expected.classid
            actual.objid = expected.objid
            assert actual == expected, f"load_procs failed for {proc_name}"



import os

import pytest
from dotenv import load_dotenv

from pgai.vectorizer.embedders import OpenAI
from pgai.vectorizer.embeddings import (
    batch_indices,
)

token_documents = [5, 1, 1, 1, 1, 1, 1, 1, 1]
string_documents = [1, 8, 2, 5, 9, 5, 6, 11]


@pytest.mark.parametrize(
    "input,batch_size,token_limit,expected,error",
    [
        (
            [],
            1,
            None,
            [],
            None,
        ),
        (
            token_documents,
            1,
            None,
            [(0, 1), (1, 2), (2, 3), (3, 4), (4, 5), (5, 6), (6, 7), (7, 8), (8, 9)],
            None,
        ),
        (
            token_documents,
            3,
            None,
            [(0, 3), (3, 6), (6, 9)],
            None,
        ),
        (
            token_documents,
            5,
            None,
            [(0, 5), (5, 9)],
            None,
        ),
        (
            token_documents,
            5,
            6,
            [(0, 2), (2, 7), (7, 9)],
            None,
        ),
        (
            token_documents,
            5,
            2,
            None,
            "chunk length 5 greater than max_tokens_per_batch 2",
        ),
        (
            string_documents,
            5,
            20,
            [(0, 4), (4, 7), (7, 8)],
            None,
        ),
    ],
)
def test_batch_indices(
    input: list[int],
    batch_size: int,
    token_limit: int | None,
    expected: list[tuple[int, int]],
    error: str,
):
    try:
        indices = batch_indices(input, batch_size, token_limit)
        assert indices == expected
    except BaseException as e:
        assert str(e) == error


@pytest.fixture
def openai_client() -> OpenAI:
    """Create an OpenAI client."""
    load_dotenv()
    client = OpenAI(
        implementation="openai",
        model="text-embedding-3-small",
        api_key_name="OPENAI_API_KEY",  # type: ignore
    )
    client.set_api_key({"OPENAI_API_KEY": os.getenv("OPENAI_API_KEY")})
    return client


async def test_openai_embeddings(openai_client: OpenAI):
    result = openai_client.embed(
        ["a"] * 800  # 800 chunks each 1 token
    )
    assert (
        len([value async for value in result][0]) == 800
    )  # One request with 800 embedding vectors


async def test_openai_embeddings_breaks_300k_tokens_but_fits_estimator(
    openai_client: OpenAI,
):
    result = openai_client.embed(
        ["ф" * 300] * 1001  # 1001 chunks each 300 token = 300300 tokens
    )
    # Estimator counts utf-8 bytes as 0.25 tokens
    # which in this case is just 150150 tokens so input goes through one request
    responses = [response async for response in result]
    assert len(responses) == 1
    assert len(responses[0]) == 1001  # One request with 1001 embedding vectors


async def test_openai_embeddings_breaks_300k_tokens_and_also_estimator(
    openai_client: OpenAI,
):
    result = openai_client.embed(
        ["中" * 1000] * 500  # 500 chunks each 1000 token = 500000 tokens
    )
    # Estimator counts utf-8 bytes as 0.25 tokens
    # zhong is 3 bytes = 0.75 tokens * 1000 * 500 = 375000 tokens
    # which should mean 2 requests
    responses = [response async for response in result]
    assert len(responses) == 2


async def test_openai_embeddings_doesnt_break_300k_tokens_but_estimator(
    openai_client: OpenAI,
):
    result = openai_client.embed(
        ["apple " * 1000] * 299  # 299 chunks each 1001 token = 299299 tokens
    )
    # Estimator counts utf-8 bytes as 0.25 tokens
    # "apple " is 6 bytes = 1.5 tokens * 1000 * 299 = 448500 tokens
    # which should mean 2 requests
    responses = [response async for response in result]
    assert len(responses) == 2
